#!/usr/bin/env python2.2

"""
Execution framework for SML/NG optimizer.  Takes care of
  * timeouts (cancelling optimization after a set period)
  * saving the original input in case we have to give up (timeout,
    or "optimized" output ends up bigger than input)

"""

import sys, os
import time
import tempfile

from sml_parser import Parser
from list_transforms import collapse_whitespace, merge_chunks, \
     merge_adjacent_whitespace
from tree_transforms import merge_simple, lift_simple_tag


# Time-consuming code is responsible for periodically checking this,
# and getting the hell out if ABORT_TIME and time.time() >= ABORT_TIME
ABORT_TIME = None


class TimeoutError (Exception):
    pass

def writenow (msg, file=None):
    if file is None:
        file = sys.stderr
    file.write(msg)
    file.flush()


def copy_data (in_file, out_file):
    size = 0
    while 1:
        data = in_file.read(8192)
        #print "got %d bytes" % len(data)
        if not data: break
        size += len(data)
        out_file.write(data)
    return size


def spew_saved_input (filename):
    file = open(filename)
    copy_data(file, sys.stdout)
    file.close()


def main ():
    args = sys.argv[1:]
    try:
        time_limit = int(args[0])
    except (IndexError, ValueError):
        sys.stderr.write("warning: no time limit given\n")
        time_limit = None

    start_time = time.time()
    if time_limit:
        ABORT_TIME = start_time + time_limit - 10 # 10 sec just to be safe


    # Save the original stdin to a temporary file.  That way, if we have
    # to give up on optimization (run out of time, ends up being
    # bigger), we can fallback on something that's no worse than the
    # original input.
    tmp_filename = tempfile.mktemp(".sml")
    tmp_file = open(tmp_filename, "w")
    #print "copying stdin to %s" % tmp_file
    input_size = copy_data(sys.stdin, tmp_file)
    tmp_file.close()
    
    try:
        try:
            writenow("initiating parse and list transformations...\n")
            in_file = open(tmp_filename)
            chunks = Parser(in_file)
            chunks = merge_chunks(chunks)
            chunks = collapse_whitespace(chunks)
            chunks = merge_adjacent_whitespace(chunks)

            #print "optimized chunks:"
            #writenow("forcing completion...\n")
            #chunks = list(chunks)
            #for c in chunks: print c

            writenow("building tree...\n")
            from tree import build_tree
            tree = build_tree(chunks)
            #print "initial tree:"
            #tree.dump()

            writenow("performing tree transformations...\n")
            while 1:
                merges = 0
                for tag in ("B", "I", "TT", "U"):
                    lift_simple_tag(tree, tag)
                    merges += merge_simple(tree)
                    #print "after lift_simple_tag(..., %s):" % tag
                    #tree.dump()
                if merges == 0:
                    break
                writenow("merged %d nodes.\n" % merges)
                
            writenow("rendering...\n")
            output = "".join(list(tree.render()))
            if len(output) < input_size:
                writenow("removed %s bytes\n" % (input_size - len(output)))
                # we did some good
                sys.stdout.write(output)
                #sys.stdout.write("\n") # end with a newline
            else:
                writenow("didn't save any bytes, spewing original\n")
                spew_saved_input(tmp_filename)
                

        except TimeoutError:
            sys.stderr.write("timed out after %d sec\n" %
                             (time.time() - start_time))
            spew_saved_input(tmp_filename)

        except KeyboardInterrupt:
            sys.exit("interrupted")

        except:
            import traceback
            (exc_type, exc_val, tb) = sys.exc_info()
            sys.stderr.write("caught %s: %s\n" % (exc_type, exc_val))
            spew_saved_input(tmp_filename)
    finally:
        try:
            import os
            os.remove(tmp_filename)
        except OSError:
            pass

# main ()

main()
