
__revision__ = "$Id: sml_parser.py,v 1.9 2001/07/27 13:31:04 nascheme Exp $"

import re, sys
from text_chunk import TextChunk as Context

tag_re = re.compile("""
< ( /? [A-Za-z0-9]+ ) > 
""", re.VERBOSE)

VALID_TAGS = ("B", "EM", "I", "PL", "S", "TT", "U",
              "0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
              "r", "g", "b", "c", "m", "y", "k", "w")

BOOLEAN_TAGS = {'B':'set_bold',
                'I':'set_italics',
                'TT':'set_tt'
                }
    
class ParseError(Exception):
    pass

class Parser:
    """
    Parses a SML/NG data stream.

    """

    def __init__ (self, file):
        """__init__ (file:string)
        Parse input coming from the file-like object 'file'.
        """
        
        self.file = file

        # Two stacks holding the currently open tags and their
        # respective contexts.
        self.tag_stack = []
        self.context_stack = [Context()]

        # Current line
        self.line = ""

    def warn (self, msg):
        "Display a warning message"
        sys.stderr.write(msg + '\n')        

    def __iter__ (self):
        # Parser is its own iterator, no need to create a new object.
        return self

    def next (self):
        """next() -> TextChunk

        Raises StopIteration when input ends.
        """
        while 1:
            result = self.process()
            if result is not None:
                return result

    def parse_all (self):
        """parse_all() -> [TextChunk]"""

        return list(self)
            
        
    def process (self):
        # Get the next line of input.
        if not self.line:
            line = self.line = self.file.readline()
            if line == "":
                raise StopIteration
        else:
            line = self.line

        # Check for tag
        m = tag_re.match(line)
        if m:
            # A tag was seen
            tag_name = m.group(1)
            self.handle_tag(tag_name)
            self.line = line[m.end():]
            return None
        
        else:
            # No tag, just text, so we read lines
            # until we see a '<'.
            lines = [line]
            while line != "" and '<' not in line:
                line = self.file.readline()
                lines.append(line)

            # The last line may have a '<' in it, or it may not, if we've
            # reached EOF.
            index = lines[-1].find('<')
            if index != -1:
                self.line = lines[-1][index:]
                lines[-1] = lines[-1][:index]
            else:
                self.line = ""

            current_text = "".join(lines)
            return self.create_chunk(current_text)

    def handle_tag (self, tag_name):
        # Figure out if this is a closing tag
        if tag_name[0] == '/':
            closing = 1
            tag_name = tag_name[1:]
        else:
            closing = 0

        # Check for a valid tag name
        if tag_name not in VALID_TAGS:
            raise ParseError, "Invalid tag: %s" % repr(tag_name)
        
        if closing:
            if not self.tag_stack:
                raise ParseError(
                    "Closing tag %r before any start tags" %  tag_name)
            
            if tag_name != self.tag_stack[-1]:
                raise ParseError(
                    "Mismatched tags: closing tag %r doesn't match opening %r"
                    % (tag_name, self.tag_stack[-1]))

            self.tag_stack.pop()
            self.context_stack.pop()
        
        else:
            # Opening tag
            self.tag_stack.append(tag_name)
            context = self.extend_context_stack()
            self.update_context(context, tag_name)
            
    def extend_context_stack (self):
        new = self.context_stack[-1].copy()
        self.context_stack.append(new)
        return new

    def update_context (self, context, tag_name):
        if BOOLEAN_TAGS.has_key(tag_name):
            meth = getattr(context, BOOLEAN_TAGS[tag_name])
            meth(1)
        elif tag_name == 'S':
            # Set the s flag and unset the emph flag
            context.set_strong(1)
            context.set_emph(0)
        elif tag_name == 'EM':
            # If already inside an <S>, <EM> is irrelevant, so ignore it.
            if not context.strong:
                context.set_emph(not context.emph)
        elif tag_name == 'PL':
            context.set_under(0)
            context.set_bold(0)
            context.set_emph(0)
            context.set_italics(0)
            context.set_strong(0)
            context.set_tt(0)
        elif tag_name == 'U':
            if context.under == 3:
                self.warn('Too many <U> tags')
            else:
                context.under += 1

        elif len(tag_name) == 1 and tag_name in '0123456789':
            context.set_size(int(tag_name))

        elif len(tag_name) == 1 and tag_name in 'rgbcmykw':
            context.set_color(tag_name)

        else:
            raise RuntimeError, "Unknown tag %r (can't happen!)" % tag_name
        
    def create_chunk (self, current_text):
        new = self.context_stack[-1].copy()
        new.set_text(current_text)
        return new
    
    
if __name__ == '__main__':
    if len(sys.argv) > 1:
        filenames = sys.argv[1:]
    else:
        filenames = ["-"]

    for filename in filenames:
        if filename == "-":
            input = sys.stdin
            filename = "(stdin)"
        else:
            input = open(filename, 'r')

        p = Parser(input)
        print '--', filename, '--'
        for result in p:
            print result
