import types

__revision__ = "$Id: text_chunk.py,v 1.9 2001/07/27 14:42:44 nascheme Exp $"

class TextChunk:

    """
    Stores one chunk of text with identical attributes.  Each element
    of the main document list is a TextChunk.

    Instance attributes:
      text : string
      bold : boolean
      emph : boolean
      italics : boolean
      strong : boolean
      tt : boolean
      under : int
      size : int
      color : None | string   # one of 'r', 'g', 'b', 'c', 'm', 'y', 'k', 'w'


    Rules:
      * parser and optimizer are allowed to modify any attributes, but
        should use set_*() methods
      * renderer should only read, and should use direct attribute access

    """

    # also behaves as leaf node in the tree
    tag = None
    children = []

    def __init__ (self):
        self.text = None
        self.bold = 0
        self.italics = 0
        self.emph = 0
        self.strong = 0
        self.tt = 0
        self.under = 0
        self.size = 0
        self.color = None

    def __str__ (self):
        return ("%r : (b=%s,i=%s,em=%s,s=%s,tt=%s,u=%s,size=%s,color=%s)"
                % (self.text, self.bold, self.italics, self.emph,
                   self.strong, self.tt, self.under, self.size, self.color))

    def __repr__ (self):
        return "<%s at %08x: %r>" % (self.__class__.__name__,
                                     id(self), self.text)

    def copy (self):
        "Return a copy of this object"
        new = TextChunk()
        new.__dict__.update(self.__dict__)
        return new

    def equals_context (self, other):
        return (self.bold == other.bold and
                self.italics == other.italics and
                self.emph == other.emph and
                self.strong == other.strong and
                self.tt == other.tt and
                self.under == other.under and
                self.size == other.size and
                self.color == other.color)

    def __eq__ (self, other):
        return self.equals_context(other) and self.text == other.text

    def equivalent_whitespace (self, other):
        """equivalent_whitespace(other : TextChunk) -> Boolean

        Return true if the two chunks can be considering equivalent
        when rendering whitespace.
        """
        
        eq = (self.tt == other.tt and
              self.under == other.under and
              self.size == other.size 
              )
        if self.under == 0:
            return eq
        else:
            return eq and self.color == other.color

    def score_context (self):
        """score_context() -> int
        Return a numeric score for the context, corresponding to how
        many bits are enabled.
        """
        score = (self.bold + self.italics + self.emph + self.strong +
                 self.tt + self.under)
        return score
        
    def append_text (self, text):
        self.text += text
    
    def set_text (self, text):
        assert type(text) is types.StringType
        self.text = text

    def set_bold (self, bold):
        assert bold in (0, 1)
        self.bold = bold

    def set_emph (self, emph):
        assert emph in (0, 1)
        self.emph = emph

    def set_italics (self, italics):
        assert italics in (0, 1)
        self.italics = italics

    def set_strong (self, strong):
        assert strong in (0, 1)
        self.strong = strong

    def set_tt (self, tt):
        assert tt in (0, 1)
        self.tt = tt

    def set_under (self, under):
        assert type(under) is types.IntType and 0 <= under <= 3
        self.under = under

    def set_size (self, size):
        assert type(size) is types.IntType and 0 <= size <= 9
        self.size = size

    def set_color (self, color):
        # XXX None ok?
        assert color in (None, 'r', 'g', 'b', 'c', 'm', 'y', 'k', 'w')
        self.color = color

    def set_plain (self):
        self.set_bold(0)
        self.set_emph(0)
        self.set_italics(0)
        self.set_strong(0)
        self.set_tt(0)
        self.set_under(0)

    def is_subset (self, other):
        """Return if 'self' is a strict subset of 'other', meaning that
        appending a number of open tags can turn 'self' into 'other'.
        """
        def xor(a,b): return (a and b) and (not a and not b)
        v = 0
        for attr in ['bold', 'emph', 'italics', 'strong', 'tt', 'under']:
            if getattr(self, attr) > getattr(other, attr):
                return 0
        if self.color is not None and other.color is None:
            return 0
        if self.size is not None and other.size is None:
            return 0
        return 1
            
    def render (self):
        return [self.text]

    # XXX ideas:
    #  naive_render() method? (no optimization, no lookahead)
        
# class TextChunk
